"""This module provides a function to parse messages from a prompt template."""

import re
from string import Formatter
from typing import Any, TypeVar

from pydantic import BaseModel

from ..message_param import BaseMessageParam
from .format_template import format_template

BaseToolT = TypeVar("BaseToolT", bound=BaseModel)


def parse_prompt_messages(
    roles: list[str], template: str, attrs: dict[str, Any]
) -> list[BaseMessageParam | Any]:
    """Returns messages parsed from the provided prompt `template`.

    Raises:
        ValueError: if `MESSAGES` keyword is used with a non-list attribute.
    """
    messages = []
    re_roles = "|".join([role.upper() for role in roles] + ["MESSAGES"])
    for match in re.finditer(rf"({re_roles}):((.|\n)+?)(?=({re_roles}):|\Z)", template):
        role = match.group(1).lower()
        if role == "messages":
            template_var = [
                var
                for _, var, _, _ in Formatter().parse(match.group(2))
                if var is not None
            ][0]
            if template_var.startswith("self"):
                if "self" not in attrs:
                    raise ValueError(
                        "MESSAGES keyword used with `self.` but `self` was not found."
                    )
                attr = getattr(attrs["self"], template_var[5:])
            else:
                attr = attrs[template_var]
            if attr is None or not isinstance(attr, list):
                raise ValueError(
                    f"MESSAGES keyword used with attribute `{template_var}`, which "
                    "is not a `list` of messages."
                )
            messages += attr
        else:
            content = format_template(match.group(2), attrs)
            if content:
                messages.append({"role": role, "content": content})
    if len(messages) == 0:
        messages.append(
            {
                "role": "user",
                "content": format_template(template, attrs),
            }
        )
    return messages



"""Utilities for the Mirascope Core OpenAI module."""

import inspect
from collections.abc import AsyncGenerator, Generator
from typing import Any, Awaitable, Callable

from google.generativeai import GenerativeModel  # type: ignore
from google.generativeai.types import (  # type: ignore
    ContentsType,
    GenerateContentResponse,
)

from ..base import BaseTool, _utils
from .call_params import GeminiCallParams
from .dynamic_config import GeminiDynamicConfig
from .tool import GeminiTool


def setup_call(
    *,
    model: str,
    client: object | None,
    fn: Callable[..., GeminiDynamicConfig | Awaitable[GeminiDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: GeminiDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: GeminiCallParams,
    extract: bool = False,
) -> tuple[
    Callable[..., GenerateContentResponse],
    str,
    list[dict[str, ContentsType]],
    list[type[GeminiTool]],
    dict[str, Any],
]:
    prompt_template, messages, tool_types, call_kwargs = _utils.setup_call(
        fn, fn_args, dynamic_config, tools, GeminiTool, call_params
    )

    gemini_messages = []
    for message in messages:
        if (role := message["role"]) == "system":
            gemini_messages += [
                {
                    "role": "user",
                    "parts": [message["content"]],
                },
                {
                    "role": "model",
                    "parts": ["Ok! I will adhere to this system message."],
                },
            ]
        else:
            gemini_messages.append({"role": role, "parts": [message["content"]]})

    if json_mode:
        generation_config = call_kwargs.get("generation_config", {})
        generation_config["response_mime_type"] = "application/json"
        call_kwargs["generation_config"] = generation_config
        gemini_messages[-1]["parts"][-1] += _utils.json_mode_content(
            tool_types[0] if tool_types else None
        )
        call_kwargs.pop("tools", None)
    elif extract:
        tool_config = call_kwargs.get("tool_config", {})
        tool_config["function_calling_config"] = {"mode": "auto"}
        call_kwargs["tool_config"] = tool_config

    if client is None:
        client = GenerativeModel(model_name=model)
    create = (
        client.generate_content_async
        if inspect.iscoroutinefunction(fn)
        else client.generate_content
    )
    call_kwargs |= {"contents": gemini_messages}

    return create, prompt_template, gemini_messages, tool_types, call_kwargs


def get_json_output(response: GenerateContentResponse, json_mode: bool) -> str:
    """Extracts the JSON output from a Gemini response."""
    content = response.candidates[0].content.parts
    if json_mode and content is not None:
        return content[0].text
    elif tool_calls := content:
        return tool_calls[0].function_call.args
    else:
        raise ValueError("No tool call or JSON object found in response.")


def handle_stream(
    stream: Generator[GenerateContentResponse, None, None],
    tool_types: list[type[GeminiTool] | Callable],
):
    """Iterator over the stream and constructs tools as they are streamed.

    Note: gemini does not currently support streaming tools.
    """
    for chunk in stream:
        yield chunk, None


async def handle_stream_async(
    stream: AsyncGenerator[GenerateContentResponse, None],
    tool_types: list[type[GeminiTool] | Callable],
):
    """Async iterator over the stream and constructs tools as they are streamed.

    Note: gemini does not currently support streaming tools.
    """
    async for chunk in stream:
        yield chunk, None


def gemini_api_calculate_cost(
    response: GenerateContentResponse, model: str
) -> float | None:
    """Calculate the cost of a Gemini API call."""
    return None



"""Internal Utilities."""

from .base_type import BaseType, is_base_type
from .convert_base_model_to_base_tool import convert_base_model_to_base_tool
from .convert_base_type_to_base_tool import convert_base_type_to_base_tool
from .convert_function_to_base_tool import convert_function_to_base_tool
from .create_base_type_with_response import create_base_type_with_response
from .default_tool_docstring import DEFAULT_TOOL_DOCSTRING
from .extract_tool_return import extract_tool_return
from .format_template import format_template
from .get_fn_args import get_fn_args
from .get_template_values import get_template_values
from .get_template_variables import get_template_variables
from .json_mode_content import json_mode_content
from .parse_prompt_messages import parse_prompt_messages
from .setup_call import setup_call
from .setup_extract_tool import setup_extract_tool

__all__ = [
    "BaseType",
    "convert_base_model_to_base_tool",
    "convert_base_type_to_base_tool",
    "convert_function_to_base_tool",
    "create_base_type_with_response",
    "DEFAULT_TOOL_DOCSTRING",
    "extract_tool_return",
    "format_template",
    "get_fn_args",
    "get_template_values",
    "get_template_variables",
    "is_base_type",
    "json_mode_content",
    "parse_prompt_messages",
    "setup_call",
    "setup_extract_tool",
]



"""Mirascope middleware utils."""

import inspect
import types
from contextlib import contextmanager
from functools import wraps
from typing import (
    Any,
    Awaitable,
    Callable,
    Generator,
    Iterable,
    ParamSpec,
    TypeVar,
)

from pydantic import BaseModel

from ..core.base import (
    BaseAsyncStream,
    BaseCallResponse,
    BaseStream,
)

_P = ParamSpec("_P")
_R = TypeVar("_R", bound=BaseCallResponse | BaseStream | BaseModel | BaseAsyncStream)
_F = TypeVar("_F", bound=Callable[..., Any])
_DecoratorType = Callable[[_F], _F]
_T = TypeVar("_T")


@contextmanager
def default_context_manager():
    yield None


def middleware(
    fn: Callable[_P, _R] | Callable[_P, Awaitable[_R]],
    *,
    custom_context_manager: Callable[
        [], Generator[_T, Any, None]
    ] = default_context_manager,
    custom_decorator: _DecoratorType | None = None,
    handle_call_response: Callable[[BaseCallResponse, _T | None], None] | None = None,
    handle_call_response_async: Callable[[BaseCallResponse, _T | None], Awaitable[None]]
    | None = None,
    handle_stream: Callable[[BaseStream, _T | None], None] | None = None,
    handle_stream_async: Callable[[BaseStream, _T | None], Awaitable[None]]
    | None = None,
    handle_base_model: Callable[[BaseModel, _T | None], None] | None = None,
    handle_base_model_async: Callable[[BaseModel, _T | None], Awaitable[None]]
    | None = None,
) -> Callable[_P, _R] | Callable[_P, Awaitable[_R]]:
    """Wraps a Mirascope function with middleware."""

    async def handlers_async(result: _R):
        if (
            isinstance(result, BaseCallResponse)
            and handle_call_response_async is not None
        ):
            with custom_context_manager() as context:
                if context is None:
                    await handle_call_response_async(result)
                else:
                    await handle_call_response_async(result, context)
        elif isinstance(result, BaseAsyncStream) and handle_stream_async is not None:
            original_aiter = result.__aiter__

            def new_aiter(self):
                async def generator():
                    with custom_context_manager() as context:
                        async for chunk, tool in original_aiter():
                            yield chunk, tool
                        if handle_stream_async is not None:
                            if context is None:
                                await handle_stream_async(result)
                            else:
                                await handle_stream_async(result, context)

                return generator()

            result.__class__.__aiter__ = types.MethodType(new_aiter, result)
        elif isinstance(result, BaseModel) and handle_base_model_async is not None:
            with custom_context_manager() as context:
                if context is None:
                    await handle_base_model_async(result)
                else:
                    await handle_base_model_async(result, context)
        elif isinstance(result, Iterable[BaseModel]):
            ...

    def inner(*args: _P.args, **kwargs: _P.kwargs) -> _R:
        result = fn(*args, **kwargs)
        if isinstance(result, BaseCallResponse) and handle_call_response is not None:
            with custom_context_manager() as context:
                if context is None:
                    handle_call_response(result)
                else:
                    handle_call_response(result, context)
        elif isinstance(result, BaseStream) and handle_stream is not None:
            original_iter = result.__iter__

            def new_iter(self):
                # Create the context manager when user iterates over the stream
                with custom_context_manager() as context:
                    for chunk, tool in original_iter():
                        yield chunk, tool
                    if handle_stream is not None:
                        if context is None:
                            handle_stream(result)
                        else:
                            handle_stream(result, context)

            result.__class__.__iter__ = new_iter
        elif isinstance(result, BaseModel) and handle_base_model is not None:
            with custom_context_manager() as context:
                if context is not None:
                    handle_base_model(result, context)
                else:
                    handle_base_model(result)
        elif isinstance(result, Iterable[BaseModel]):
            ...
        return result

    async def inner_async(*args: _P.args, **kwargs: _P.kwargs) -> Awaitable[_R]:
        result = await fn(*args, **kwargs)
        await handlers_async(result)
        return result

    inner_fn = inner_async if inspect.iscoroutinefunction(fn) else inner
    decorated_fn = custom_decorator(inner_fn) if custom_decorator else inner_fn
    return wraps(fn)(decorated_fn)



"""Utilities for the Mirascope Core OpenAI module."""

import inspect
from collections.abc import AsyncGenerator, Generator
from typing import Any, Awaitable, Callable

from openai import AsyncOpenAI, OpenAI
from openai._base_client import BaseClient
from openai.types.chat import (
    ChatCompletion,
    ChatCompletionChunk,
    ChatCompletionMessageParam,
    ChatCompletionMessageToolCall,
)
from openai.types.chat.chat_completion_message_tool_call import Function

from ..base import BaseTool, _utils
from .call_params import OpenAICallParams
from .call_response_chunk import OpenAICallResponseChunk
from .dyanmic_config import OpenAIDynamicConfig
from .tool import OpenAITool


def setup_call(
    *,
    model: str,
    client: BaseClient | None,
    fn: Callable[..., OpenAIDynamicConfig | Awaitable[OpenAIDynamicConfig]],
    fn_args: dict[str, any],
    dynamic_config: OpenAIDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: OpenAICallParams,
    extract: bool,
) -> tuple[
    Callable[..., ChatCompletion],
    str,
    list[dict[str, ChatCompletionMessageParam]],
    list[type[OpenAITool]],
    dict[str, Any],
]:
    prompt_template, messages, tool_types, call_kwargs = _utils.setup_call(
        fn, fn_args, dynamic_config, tools, OpenAITool, call_params
    )
    if client is None:
        client = AsyncOpenAI() if inspect.iscoroutinefunction(fn) else OpenAI()
    create = client.chat.completions.create
    call_kwargs |= {"model": model, "messages": messages}
    if json_mode:
        call_kwargs["response_format"] = {"type": "json_object"}
        messages[-1]["content"] += _utils.json_mode_content(tools[0] if tools else None)
        call_kwargs.pop("tools", None)
    elif extract:
        call_kwargs["tool_choice"] = "required"

    return create, prompt_template, messages, tool_types, call_kwargs


def get_json_output(response: ChatCompletion, json_mode: bool) -> str:
    """Get the JSON output from a completion response."""
    if json_mode and (content := response.choices[0].message.content):
        return content
    elif tool_calls := response.choices[0].message.tool_calls:
        return tool_calls[0].function.arguments
    else:
        raise ValueError("No tool call or JSON object found in response.")


def _handle_chunk(
    chunk: ChatCompletionChunk,
    current_tool_call: ChatCompletionMessageToolCall,
    current_tool_type: type[OpenAITool] | None,
    tool_types: list[type[OpenAITool]] | None,
) -> tuple[
    OpenAITool | None,
    ChatCompletionMessageToolCall,
    type[OpenAITool] | None,
]:
    """Handles a chunk of the stream."""
    if not tool_types or not (tool_calls := chunk.choices[0].delta.tool_calls):
        return None, current_tool_call, current_tool_type

    tool_call = tool_calls[0]
    # Reset on new tool
    if tool_call.id and tool_call.function is not None:
        previous_tool_call = current_tool_call.model_copy()
        previous_tool_type = current_tool_type
        current_tool_call = ChatCompletionMessageToolCall(
            id=tool_call.id,
            function=Function(
                arguments="",
                name=tool_call.function.name if tool_call.function.name else "",
            ),
            type="function",
        )
        current_tool_type = None
        for tool_type in tool_types:
            if tool_type._name() == tool_call.function.name:
                current_tool_type = tool_type
                break
        if current_tool_type is None:
            raise RuntimeError(
                f"Unknown tool type in stream: {tool_call.function.name}"
            )
        if previous_tool_call.id and previous_tool_type is not None:
            return (
                previous_tool_type.from_tool_call(previous_tool_call),
                current_tool_call,
                current_tool_type,
            )

    # Update arguments with each chunk
    if tool_call.function and tool_call.function.arguments:
        current_tool_call.function.arguments += tool_call.function.arguments

    return None, current_tool_call, current_tool_type


def handle_stream(
    stream: Generator[ChatCompletionChunk, None, None],
    tool_types: list[type[OpenAITool]] | None,
) -> Generator[tuple[OpenAICallResponseChunk, OpenAITool | None], None, None]:
    """Iterates over the stream and constructs tools as they are streamed."""
    current_tool_call = ChatCompletionMessageToolCall(
        id="", function=Function(arguments="", name=""), type="function"
    )
    current_tool_type = None
    for chunk in stream:
        if not tool_types or not chunk.choices[0].delta.tool_calls:
            if current_tool_type:
                yield (
                    chunk,
                    current_tool_type.from_tool_call(current_tool_call),
                )
                current_tool_type = None
            else:
                yield chunk, None
        tool, current_tool_call, current_tool_type = _handle_chunk(
            chunk,
            current_tool_call,
            current_tool_type,
            tool_types,
        )
        if tool is not None:
            yield chunk, tool


async def handle_stream_async(
    stream: AsyncGenerator[ChatCompletionChunk, None],
    tool_types: list[type[OpenAITool]] | None,
):
    """Async iterates over the stream and constructs tools as they are streamed."""
    current_tool_call = ChatCompletionMessageToolCall(
        id="", function=Function(arguments="", name=""), type="function"
    )
    current_tool_type = None
    async for chunk in stream:
        if not tool_types or not chunk.choices[0].delta.tool_calls:
            if current_tool_type:
                yield (
                    chunk,
                    current_tool_type.from_tool_call(current_tool_call),
                )
                current_tool_type = None
            else:
                yield chunk, None
        tool, current_tool_call, current_tool_type = _handle_chunk(
            chunk,
            current_tool_call,
            current_tool_type,
            tool_types,
        )
        if tool is not None:
            yield chunk, tool


def openai_api_calculate_cost(
    response: ChatCompletion, model="gpt-3.5-turbo-16k"
) -> float | None:
    """Calculate the cost of a completion using the OpenAI API.

    https://openai.com/pricing

    Model                   Input               Output
    gpt-4o                  $5.00 / 1M tokens   $15.00 / 1M tokens
    gpt-4o-2024-05-13       $5.00 / 1M tokens   $15.00 / 1M tokens
    gpt-4-turbo             $10.00 / 1M tokens  $30.00 / 1M tokens
    gpt-4-turbo-2024-04-09  $10.00 / 1M tokens  $30.00 / 1M tokens
    gpt-3.5-turbo-0125	    $0.50 / 1M tokens	$1.50 / 1M tokens
    gpt-3.5-turbo-1106	    $1.00 / 1M tokens	$2.00 / 1M tokens
    gpt-4-1106-preview	    $10.00 / 1M tokens 	$30.00 / 1M tokens
    gpt-4	                $30.00 / 1M tokens	$60.00 / 1M tokens
    text-embedding-3-small	$0.02 / 1M tokens
    text-embedding-3-large	$0.13 / 1M tokens
    text-embedding-ada-0002	$0.10 / 1M tokens
    """
    pricing = {
        "gpt-4o": {
            "prompt": 0.000_005,
            "completion": 0.000_015,
        },
        "gpt-4o-2024-05-13": {
            "prompt": 0.000_005,
            "completion": 0.000_015,
        },
        "gpt-4-turbo": {
            "prompt": 0.000_01,
            "completion": 0.000_03,
        },
        "gpt-4-turbo-2024-04-09": {
            "prompt": 0.000_01,
            "completion": 0.000_03,
        },
        "gpt-3.5-turbo-0125": {
            "prompt": 0.000_000_5,
            "completion": 0.000_001_5,
        },
        "gpt-3.5-turbo-1106": {
            "prompt": 0.000_001,
            "completion": 0.000_002,
        },
        "gpt-4-1106-preview": {
            "prompt": 0.000_01,
            "completion": 0.000_03,
        },
        "gpt-4": {
            "prompt": 0.000_003,
            "completion": 0.000_006,
        },
        "gpt-3.5-turbo-4k": {
            "prompt": 0.000_015,
            "completion": 0.000_02,
        },
        "gpt-3.5-turbo-16k": {
            "prompt": 0.000_003,
            "completion": 0.000_004,
        },
        "gpt-4-8k": {
            "prompt": 0.000_003,
            "completion": 0.000_006,
        },
        "gpt-4-32k": {
            "prompt": 0.000_006,
            "completion": 0.000_012,
        },
        "text-embedding-3-small": {
            "prompt": 0.000_000_02,
            "completion": 0.000_000_02,
        },
        "text-embedding-ada-002": {
            "prompt": 0.000_000_1,
            "completion": 0.000_000_1,
        },
        "text-embedding-3-large": {
            "prompt": 0.000_000_13,
            "completion": 0.000_000_13,
        },
    }
    if (usage := response.usage) is None:
        return None
    try:
        model = response.model if response.model else model
        model_pricing = pricing[model]
    except KeyError:
        return None

    prompt_cost = usage.prompt_tokens * model_pricing["prompt"]
    completion_cost = usage.completion_tokens * model_pricing["completion"]
    total_cost = prompt_cost + completion_cost

    return total_cost



"""This module contains the OpenAI `structured_stream_decorator` function."""

from collections.abc import Generator
from functools import wraps
from typing import Callable, Generic, Iterable, ParamSpec, TypeVar

from openai import OpenAI
from openai._base_client import BaseClient
from openai.types.chat import ChatCompletionChunk
from pydantic import BaseModel

from ..base import BaseStructuredStream, _utils

# from ._utils import setup_extract
from .call_params import OpenAICallParams
from .dyanmic_config import OpenAIDynamicConfig
from .tool import OpenAITool

_P = ParamSpec("_P")
_ResponseModelT = TypeVar("_ResponseModelT", bound=BaseModel | _utils.BaseType)


class OpenAIStructuredStream(
    Generic[_ResponseModelT],
    BaseStructuredStream[ChatCompletionChunk, _ResponseModelT],
):
    """A class for streaming structured outputs from OpenAI's API."""

    def __iter__(self) -> Generator[_ResponseModelT, None, None]:
        """Iterates over the stream and extracts structured outputs."""
        json_output = ""
        for chunk in self.stream:
            if (
                self.json_mode
                and (content := chunk.choices[0].delta.content) is not None
            ):
                json_output += content
            elif (
                (tool_calls := chunk.choices[0].delta.tool_calls)
                and (function := tool_calls[0].function)
                and (arguments := function.arguments)
            ):
                json_output += arguments
            else:
                ValueError("No tool call or JSON object found in response.")
            if json_output:
                yield _utils.extract_tool_return(self.response_model, json_output, True)
        yield _utils.extract_tool_return(self.response_model, json_output, False)


def structured_stream_decorator(
    fn: Callable[_P, OpenAIDynamicConfig],
    model: str,
    response_model: type[_ResponseModelT],
    client: BaseClient | None,
    call_params: OpenAICallParams,
) -> Callable[_P, Iterable[_ResponseModelT]]:
    assert response_model is not None
    tool = _utils.setup_extract_tool(response_model, OpenAITool)

    @wraps(fn)
    def inner(*args: _P.args, **kwargs: _P.kwargs) -> Iterable[_ResponseModelT]:
        # assert response_model is not None
        # fn_args = _utils.get_fn_args(fn, args, kwargs)
        # fn_return = fn(*args, **kwargs)
        # json_mode, messages, call_kwargs = setup_extract(
        #     fn, fn_args, fn_return, tool, call_params
        # )
        # _client = client or OpenAI()
        # return OpenAIStructuredStream(
        #     stream=(
        #         chunk
        #         for chunk in _client.chat.completions.create(
        #             model=model, stream=True, messages=messages, **call_kwargs
        #         )
        #     ),
        #     response_model=response_model,
        #     json_mode=json_mode,
        # )
        ...

    return inner



"""This module provides a function to get the variables in a template string."""

from string import Formatter


def get_template_variables(template: str) -> list[str]:
    """Returns the variables in the given template string.

    Args:
        template: The template string to parse.

    Returns:
        The variables in the template string.
    """
    return [var for _, var, _, _ in Formatter().parse(template) if var is not None]



"""The `OpenAITool` class for easy tool usage with OpenAI LLM calls."""

from __future__ import annotations

import copy

from anthropic.types import ToolParam, ToolUseBlock

from ..base import BaseTool


class AnthropicTool(BaseTool):
    """A class for defining tools for Anthropic LLM calls."""

    tool_call: ToolUseBlock

    @classmethod
    def tool_schema(cls) -> ToolParam:
        """Constructs a `ToolParam` tool schema from the `BaseModel` schema defined."""
        model_schema = cls.model_json_schema()
        model_schema.pop("title", None)
        model_schema.pop("description", None)

        return ToolParam(
            input_schema=model_schema,
            name=cls._name(),
            description=cls._description(),
        )

    @classmethod
    def from_tool_call(cls, tool_call: ToolUseBlock) -> AnthropicTool:
        """Constructs an `AnthropicTool` instance from a `tool_call`."""
        model_json = copy.deepcopy(tool_call.input)
        model_json["tool_call"] = tool_call.model_dump()
        return cls.model_validate(model_json)



"""Integrations with third party libraries."""

from contextlib import suppress

with suppress(ImportError):
    from . import tenacity as tenacity

with suppress(ImportError):
    from . import logfire as logfire

__all__ = ["tenacity", "logfire"]



"""Mirascope x Logfire Integration."""

from contextlib import contextmanager
from typing import (
    Any,
    Awaitable,
    Callable,
    Generator,
    ParamSpec,
    TypeVar,
    overload,
)

import logfire
from pydantic import BaseModel

from ..core.base import BaseAsyncStream, BaseCallResponse, BaseStream
from .utils import middleware

_P = ParamSpec("_P")
_R = TypeVar("_R", bound=BaseCallResponse | BaseStream | BaseModel | BaseAsyncStream)


@overload
def with_logfire(fn: Callable[_P, _R]) -> Callable[_P, _R]: ...


@overload
def with_logfire(fn: Callable[_P, Awaitable[_R]]) -> Callable[_P, Awaitable[_R]]: ...


def with_logfire(
    fn: Callable[_P, _R] | Callable[_P, Awaitable[_R]],
) -> Callable[_P, _R] | Callable[_P, Awaitable[_R]]:
    """Wraps a Mirascope function with Logfire tracing."""

    @contextmanager
    def custom_context_manager() -> Generator[logfire.LogfireSpan, Any, None]:
        with logfire.with_settings(
            custom_scope_suffix="mirascope", tags=fn.__annotations__.get("tags", [])
        ).span(fn.__name__) as logfire_span:
            yield logfire_span

    def handle_call_response(
        result: BaseCallResponse, logfire_span: logfire.LogfireSpan | None
    ):
        if logfire_span is None:
            return
        output: dict[str, Any] = {}
        span_data = {
            "async": False,
            "call_params": result.call_params,
            "model": result.model,
            "provider": result.provider,
            "prompt_template": result.prompt_template,
            "template_variables": result.fn_args,
            "output": output,
            "messages": result.messages,
            "response_data": result.response,
        }

        if tools := result.tools:
            tool_calls = [
                {
                    "function": {
                        "arguments": tool.model_dump_json(exclude={"tool_call"}),
                        "name": tool.name(),
                    }
                }
                for tool in tools
            ]
            output["tool_calls"] = tool_calls
        if cost := result.cost:
            output["cost"] = cost
        if input_tokens := result.input_tokens:
            output["input_tokens"] = input_tokens
        if output_tokens := result.output_tokens:
            output["output_tokens"] = output_tokens
        if content := result.content:
            output["content"] = content
        logfire_span.set_attributes(span_data)

    def handle_stream(
        stream: BaseStream,
        logfire_span: logfire.LogfireSpan | None,
    ):
        output: dict[str, Any] = {
            "cost": stream.cost,
            "input_tokens": stream.input_tokens,
            "output_tokens": stream.output_tokens,
            "content": stream.message_param.get("content", None),
        }
        span_data = {
            "async": False,
            "output": output,
            "messages": [stream.user_message_param],
            "call_params": stream.call_params,
            "model": stream.model,
            "provider": stream.provider,
            "prompt_template": stream.prompt_template,
            "template_variables": stream.fn_args,
        }
        if logfire_span is not None:
            logfire_span.set_attributes(span_data)

    def handle_base_model(result: BaseModel, logfire_span: logfire.LogfireSpan | None):
        response: BaseCallResponse = result._response
        # handle_call_response(response, logfire_span)
        output: dict[str, Any] = {"response_model": result}
        # if cost := response.cost:
        #     output["cost"] = cost
        # if input_tokens := response.input_tokens:
        #     output["input_tokens"] = input_tokens
        # if output_tokens := response.output_tokens:
        #     output["output_tokens"] = output_tokens
        # if content := response.content:
        #     output["content"] = content
        logfire_span.set_attribute("output", output)

    async def handle_call_response_async(
        result: BaseCallResponse, logfire_span: logfire.LogfireSpan | None
    ):
        if logfire_span is None:
            return
        output: dict[str, Any] = {}
        span_data = {
            "async": True,
            "call_params": result.call_params,
            "model": result.model,
            "provider": result.provider,
            "prompt_template": result.prompt_template,
            "template_variables": result.fn_args,
            "output": output,
            "messages": result.messages,
            "response_data": result.response,
        }

        if tools := result.tools:
            tool_calls = [
                {
                    "function": {
                        "arguments": tool.model_dump_json(exclude={"tool_call"}),
                        "name": tool.name(),
                    }
                }
                for tool in tools
            ]
            output["tool_calls"] = tool_calls
        if cost := result.cost:
            output["cost"] = cost
        if input_tokens := result.input_tokens:
            output["input_tokens"] = input_tokens
        if output_tokens := result.output_tokens:
            output["output_tokens"] = output_tokens
        if content := result.content:
            output["content"] = content
        logfire_span.set_attributes(span_data)

    async def handle_stream_async(
        stream: BaseStream,
        logfire_span: logfire.LogfireSpan | None,
    ):
        output: dict[str, Any] = {
            "cost": stream.cost,
            "input_tokens": stream.input_tokens,
            "output_tokens": stream.output_tokens,
            "content": stream.message_param.get("content", None),
        }
        span_data = {
            "async": True,
            "output": output,
            "messages": [stream.user_message_param],
            "call_params": stream.call_params,
            "model": stream.model,
            "provider": stream.provider,
            "prompt_template": stream.prompt_template,
            "template_variables": stream.fn_args,
        }
        if logfire_span is not None:
            logfire_span.set_attributes(span_data)

    return middleware(
        fn,
        custom_context_manager=custom_context_manager,
        handle_call_response=handle_call_response,
        handle_call_response_async=handle_call_response_async,
        handle_stream=handle_stream,
        handle_stream_async=handle_stream_async,
        handle_base_model=handle_base_model,
        # handle_base_model_async=handle_base_model_async,
    )



"""This module contains the OpenAI `structured_stream_async_decorator` function."""

from collections.abc import AsyncGenerator
from functools import wraps
from typing import (
    AsyncIterable,
    Awaitable,
    Callable,
    Generic,
    ParamSpec,
    TypeVar,
)

from openai import AsyncOpenAI
from openai.types.chat import ChatCompletionChunk
from pydantic import BaseModel

from ..base import BaseAsyncStructuredStream, _utils

# from ._utils import setup_extract
from .call_params import OpenAICallParams
from .dyanmic_config import OpenAIDynamicConfig
from .tool import OpenAITool

_P = ParamSpec("_P")
_ResponseModelT = TypeVar("_ResponseModelT", bound=BaseModel | _utils.BaseType)


class OpenAIAsyncStructuredStream(
    Generic[_ResponseModelT],
    BaseAsyncStructuredStream[ChatCompletionChunk, _ResponseModelT],
):
    """A class for async streaming structured outputs from OpenAI's API."""

    def __aiter__(self) -> AsyncGenerator[_ResponseModelT, None]:
        """Iterates over the stream and extracts structured outputs."""

        async def generator():
            nonlocal self
            json_output = ""
            async for chunk in self.stream:
                if (
                    self.json_mode
                    and (content := chunk.choices[0].delta.content) is not None
                ):
                    json_output += content
                elif (
                    (tool_calls := chunk.choices[0].delta.tool_calls)
                    and (function := tool_calls[0].function)
                    and (arguments := function.arguments)
                ):
                    json_output += arguments
                else:
                    ValueError("No tool call or JSON object found in response.")
                if json_output:
                    yield _utils.extract_tool_return(
                        self.response_model, json_output, True
                    )
            yield _utils.extract_tool_return(self.response_model, json_output, False)

        return generator()


def structured_stream_async_decorator(
    fn: Callable[_P, Awaitable[OpenAIDynamicConfig]],
    model: str,
    response_model: type[_ResponseModelT],
    call_params: OpenAICallParams,
) -> Callable[_P, Awaitable[AsyncIterable[_ResponseModelT]]]:
    assert response_model is not None
    tool = _utils.setup_extract_tool(response_model, OpenAITool)

    @wraps(fn)
    async def inner_async(
        *args: _P.args, **kwargs: _P.kwargs
    ) -> AsyncIterable[_ResponseModelT]:
        # assert response_model is not None
        # fn_args = _utils.get_fn_args(fn, args, kwargs)
        # fn_return = await fn(*args, **kwargs)
        # json_mode, messages, call_kwargs = setup_extract(
        #     fn, fn_args, fn_return, tool, call_params
        # )
        # client = AsyncOpenAI()
        # return OpenAIAsyncStructuredStream(
        #     stream=(
        #         chunk
        #         async for chunk in await client.chat.completions.create(
        #             model=model, stream=True, messages=messages, **call_kwargs
        #         )
        #     ),
        #     response_model=response_model,
        #     json_mode=json_mode,
        # )
        ...

    return inner_async



"""The Mirascope OpenAI Module."""

from .call import openai_call
from .call import openai_call as call
from .call_async import openai_call_async
from .call_async import openai_call_async as call_async
from .call_params import OpenAICallParams
from .call_response import OpenAICallResponse
from .call_response_chunk import OpenAICallResponseChunk
from .dyanmic_config import OpenAIDynamicConfig
from .tool import OpenAITool

__all__ = [
    "call",
    "call_async",
    "OpenAIDynamicConfig",
    "OpenAICallParams",
    "OpenAICallResponse",
    "OpenAICallResponseChunk",
    "OpenAITool",
    "openai_call",
    "openai_call_async",
]



"""This module contains the `OpenAICallResponseChunk` class."""

from openai.types.chat import ChatCompletionChunk
from openai.types.chat.chat_completion_chunk import (
    Choice,
    ChoiceDelta,
    ChoiceDeltaToolCall,
)
from openai.types.completion_usage import CompletionUsage

from ..base import BaseCallResponseChunk


class OpenAICallResponseChunk(BaseCallResponseChunk[ChatCompletionChunk]):
    '''A convenience wrapper around the OpenAI `ChatCompletionChunk` streamed chunks.

    When calling the OpenAI API using a function decorated with `openai_call` and
    `stream` set to `True`, the stream will contain `OpenAIResponseChunk` instances with
    properties that allow for more convenient access to commonly used attributes.

    Example:

    ```python
    from mirascope.core.openai import openai_call

    @openai_call(model="gpt-4o", stream=True)
    def recommend_book(genre: str):
        """Recommend a {genre} book."""

    stream = recommend_book("fantasy")  # response is an `OpenAIStream`
    for chunk in stream:
        print(chunk.content, end="", flush=True)
    #> Sure! I would recommend...
    ```
    '''

    @property
    def choices(self) -> list[Choice]:
        """Returns the array of chat completion choices."""
        return self.chunk.choices

    @property
    def choice(self) -> Choice:
        """Returns the 0th choice."""
        return self.chunk.choices[0]

    @property
    def delta(self) -> ChoiceDelta | None:
        """Returns the delta for the 0th choice."""
        if self.chunk.choices:
            return self.chunk.choices[0].delta
        return None

    @property
    def content(self) -> str:
        """Returns the content for the 0th choice delta."""
        return (
            self.delta.content if self.delta is not None and self.delta.content else ""
        )

    @property
    def model(self) -> str:
        """Returns the name of the response model."""
        return self.chunk.model

    @property
    def id(self) -> str:
        """Returns the id of the response."""
        return self.chunk.id

    @property
    def finish_reasons(self) -> list[str]:
        """Returns the finish reasons of the response."""
        return [str(choice.finish_reason) for choice in self.chunk.choices]

    @property
    def tool_calls(self) -> list[ChoiceDeltaToolCall] | None:
        """Returns the partial tool calls for the 0th choice message.

        The first `list[ChoiceDeltaToolCall]` will contain the name of the tool and
        index, and subsequent `list[ChoiceDeltaToolCall]`s will contain the arguments
        which will be strings that need to be concatenated with future
        `list[ChoiceDeltaToolCall]`s to form a complete JSON tool calls. The last
        `list[ChoiceDeltaToolCall]` will be None indicating end of stream.
        """
        if self.delta:
            return self.delta.tool_calls
        return None

    @property
    def usage(self) -> CompletionUsage | None:
        """Returns the usage of the chat completion."""
        if self.chunk.usage:
            return self.chunk.usage
        return None

    @property
    def input_tokens(self) -> int | None:
        """Returns the number of input tokens."""
        if self.usage:
            return self.usage.prompt_tokens
        return None

    @property
    def output_tokens(self) -> int | None:
        """Returns the number of output tokens."""
        if self.usage:
            return self.usage.completion_tokens
        return None



"""The `call_async_factory` method for generating provider specific call decorators."""

from collections.abc import AsyncGenerator, Generator
from functools import partial
from typing import (
    Any,
    AsyncIterable,
    Awaitable,
    Callable,
    Literal,
    NoReturn,
    ParamSpec,
    TypeVar,
    Unpack,
    overload,
)

from pydantic import BaseModel

from ._create import create_factory
from ._extract import extract_factory
from ._stream import BaseStream, stream_factory
from ._utils import BaseType
from .call_response import BaseCallResponse
from .call_response_chunk import BaseCallResponseChunk
from .dynamic_config import BaseDynamicConfig
from .tool import BaseTool

_BaseCallResponseT = TypeVar("_BaseCallResponseT", bound=BaseCallResponse)
_BaseCallResponseChunkT = TypeVar(
    "_BaseCallResponseChunkT", bound=BaseCallResponseChunk
)
_ResponseModelT = TypeVar("_ResponseModelT", bound=BaseModel | BaseType)
_ExtractModelT = TypeVar("_ExtractModelT", bound=BaseModel | BaseType)
_ParsedOutputT = TypeVar("_ParsedOutputT")
_BaseCallParamsT = TypeVar("_BaseCallParamsT", bound=BaseModel)
_BaseDynamicConfigT = TypeVar("_BaseDynamicConfigT", bound=BaseDynamicConfig)
_BaseStreamT = TypeVar("_BaseStreamT", bound=BaseStream)
_BaseClientT = TypeVar("_BaseClientT", bound=object)
_BaseToolT = TypeVar("_BaseToolT", bound=BaseTool)
_ResponseT = TypeVar("_ResponseT")
_AssistantMessageParamT = TypeVar("_AssistantMessageParamT")
_P = ParamSpec("_P")


def call_async_factory(
    *,
    TCallResponse: type[_BaseCallResponseT],
    TCallResponseChunk: type[_BaseCallResponseChunkT],
    TCallParams: type[_BaseCallParamsT],
    TDynamicConfig: type[_BaseDynamicConfigT],
    TStream: type[_BaseStreamT],
    TMessageParamType: type[_AssistantMessageParamT],
    TToolType: type[_BaseToolT],
    setup_call: Callable[
        [
            str,
            _BaseClientT,
            Callable[_P, _BaseDynamicConfigT | Awaitable[_BaseDynamicConfigT]],
            dict[str, Any],
            _BaseDynamicConfigT,
            list[type[BaseTool] | Callable] | None,
            _BaseCallParamsT,
            bool,
        ],
        tuple[
            Callable[..., _ResponseT],
            str,
            list[dict[str, Any]],
            list[type[BaseTool]],
            dict[str, Any],
        ],
    ],
    get_json_output: Callable[[_ResponseT, bool], str],
    handle_stream: Callable[
        [Generator[_BaseCallResponseChunkT, None, None], list[type[_BaseToolT]]],
        Generator[tuple[_BaseCallResponseChunkT, _BaseToolT], None, None],
    ],
    handle_stream_async: Callable[
        [AsyncGenerator[_BaseCallResponseChunkT, None], list[type[_BaseToolT]]],
        Awaitable[AsyncGenerator[tuple[_BaseCallResponseChunkT, _BaseToolT], None]],
    ],
    calculate_cost: Callable[[_ResponseT, str], float],
    structured_stream_async_decorator: Callable[
        [
            Callable[_P, Awaitable[_BaseDynamicConfigT]],
            str,
            type[_ExtractModelT],
            _BaseCallParamsT,
        ],
        Callable[_P, Awaitable[AsyncIterable[_ExtractModelT]]],
    ],
):
    @overload
    def base_call(
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: None = None,
        json_mode: bool = False,
        client: _BaseClientT | None = None,
        **call_params: Unpack[TCallParams],
    ) -> Callable[
        [Callable[_P, Awaitable[TDynamicConfig]]],
        Callable[_P, Awaitable[TCallResponse]],
    ]: ...  # pragma: no cover

    @overload
    def base_call(
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: Callable[[TCallResponse], _ParsedOutputT],
        json_mode: bool = False,
        client: _BaseClientT | None = None,
        **call_params: Unpack[TCallParams],
    ) -> Callable[
        [Callable[_P, Awaitable[TDynamicConfig]]],
        Callable[_P, Awaitable[_ParsedOutputT]],
    ]: ...  # pragma: no cover

    @overload
    def base_call(
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: Callable[[TCallResponseChunk], _ParsedOutputT],
        json_mode: bool = False,
        client: _BaseClientT | None = None,
        **call_params: Unpack[TCallParams],
    ) -> NoReturn: ...  # pragma: no cover

    @overload
    def base_call(
        model: str,
        *,
        stream: Literal[True] = True,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: None = None,
        json_mode: bool = False,
        client: _BaseClientT | None = None,
        **call_params: Unpack[TCallParams],
    ) -> Callable[
        [Callable[_P, Awaitable[TDynamicConfig]]],
        Callable[_P, Awaitable[TStream]],
    ]: ...  # pragma: no cover

    @overload
    def base_call(
        model: str,
        *,
        stream: Literal[True] = True,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: Callable[[TCallResponseChunk], _ParsedOutputT],
        json_mode: bool = False,
        client: _BaseClientT | None = None,
        **call_params: Unpack[TCallParams],
    ) -> NoReturn: ...  # pragma: no cover

    @overload
    def base_call(
        model: str,
        *,
        stream: Literal[True] = True,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: None = None,
        output_parser: Callable[[TCallResponse], _ParsedOutputT],
        json_mode: bool = False,
        client: _BaseClientT | None = None,
        **call_params: Unpack[TCallParams],
    ) -> NoReturn: ...  # pragma: no cover

    @overload
    def base_call(
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: type[_ResponseModelT],
        output_parser: None = None,
        json_mode: bool = False,
        client: _BaseClientT | None = None,
        **call_params: Unpack[TCallParams],
    ) -> Callable[
        [Callable[_P, Awaitable[TDynamicConfig]]],
        Callable[_P, Awaitable[_ResponseModelT]],
    ]: ...  # pragma: no cover

    @overload
    def base_call(
        model: str,
        *,
        stream: Literal[False] = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: type[_ResponseModelT],
        output_parser: Callable[[TCallResponse], _ParsedOutputT]
        | Callable[[TCallResponseChunk], _ParsedOutputT],
        json_mode: bool = False,
        client: _BaseClientT | None = None,
        **call_params: Unpack[TCallParams],
    ) -> NoReturn: ...  # pragma: no cover

    @overload
    def base_call(
        model: str,
        *,
        stream: Literal[True],
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: type[_ResponseModelT],
        output_parser: None = None,
        json_mode: bool = False,
        client: _BaseClientT | None = None,
        **call_params: Unpack[TCallParams],
    ) -> Callable[
        [Callable[_P, Awaitable[TDynamicConfig]]],
        Callable[_P, Awaitable[AsyncIterable[_ResponseModelT]]],
    ]: ...  # pragma: no cover

    @overload
    def base_call(
        model: str,
        *,
        stream: Literal[True],
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: type[_ResponseModelT],
        output_parser: Callable[[TCallResponse], _ParsedOutputT]
        | Callable[[TCallResponseChunk], _ParsedOutputT],
        json_mode: bool = False,
        client: _BaseClientT | None = None,
        **call_params: Unpack[TCallParams],
    ) -> NoReturn: ...  # pragma: no cover

    def base_call(
        model: str,
        *,
        stream: bool = False,
        tools: list[type[BaseTool] | Callable] | None = None,
        response_model: type[_ResponseModelT] | None = None,
        output_parser: Callable[[TCallResponse], _ParsedOutputT]
        | Callable[[TCallResponseChunk], _ParsedOutputT]
        | None = None,
        json_mode: bool = False,
        client: _BaseClientT | None = None,
        **call_params: Unpack[TCallParams],
    ) -> Callable[
        [Callable[_P, Awaitable[TDynamicConfig]]],
        Callable[
            _P,
            Awaitable[TCallResponse]
            | Awaitable[_ParsedOutputT]
            | Awaitable[TStream]
            | Awaitable[_ResponseModelT]
            | Awaitable[AsyncIterable[_ResponseModelT]],
        ],
    ]:
        if stream and output_parser:
            raise ValueError("Cannot use `output_parser` with `stream=True`.")
        if response_model and output_parser:
            raise ValueError("Cannot use both `response_model` and `output_parser`.")

        if response_model:
            if stream:
                return partial(
                    structured_stream_async_decorator,
                    model=model,
                    response_model=response_model,
                    call_params=call_params,
                )
            else:
                return partial(
                    extract_factory(
                        TCallResponse=TCallResponse,
                        TToolType=TToolType,
                        setup_call=setup_call,
                        get_json_output=get_json_output,
                        calculate_cost=calculate_cost,
                    ),
                    model=model,
                    response_model=response_model,
                    json_mode=json_mode,
                    client=client,
                    call_params=call_params,
                )
        if stream:
            return partial(
                stream_factory(
                    TCallResponseChunk=TCallResponseChunk,
                    TStream=TStream,
                    TMessageParamType=TMessageParamType,
                    setup_call=setup_call,
                    handle_stream=handle_stream,
                    handle_stream_async=handle_stream_async,
                ),
                model=model,
                tools=tools,
                json_mode=json_mode,
                client=client,
                call_params=call_params,
            )
        return partial(
            create_factory(
                TCallResponse=TCallResponse,
                setup_call=setup_call,
                calculate_cost=calculate_cost,
            ),
            model=model,
            tools=tools,
            output_parser=output_parser,
            json_mode=json_mode,
            client=client,
            call_params=call_params,
        )

    return base_call



"""This module contains the function to create a base type with response."""

import types

from .base_type import BaseType


def create_base_type_with_response(type_: type[BaseType]):
    class_name = f"{type_.__name__.capitalize()}WithResponse"

    # Define the __new__ method
    def __new__(cls, *args, **kwargs):
        instance = super(cls, cls).__new__(cls, *args, **kwargs)
        instance._response = None
        return instance

    # Check if the type is a generic type
    if hasattr(type_, "__origin__"):
        base_type = type_.__origin__
    else:
        base_type = type_

    BaseTypeWithResponse = types.new_class(
        class_name,
        (base_type,),  # Base classes
        exec_body=lambda ns: ns.update(
            {
                "__new__": __new__,
            }
        ),
    )

    return BaseTypeWithResponse



"""Utilities for the Mirascope Core Anthropic module."""

import inspect
import json
from collections.abc import AsyncGenerator, Generator
from typing import Any, Awaitable, Callable

from anthropic import Anthropic, AsyncAnthropic
from anthropic._base_client import BaseClient
from anthropic.types import (
    Message,
    MessageParam,
    MessageStreamEvent,
    TextBlock,
    ToolUseBlock,
)

from ..base import BaseTool, _utils
from .call_params import AnthropicCallParams
from .call_response_chunk import AnthropicCallResponseChunk
from .dynamic_config import AnthropicDynamicConfig
from .tool import AnthropicTool


def setup_call(
    *,
    model: str,
    client: BaseClient | None,
    fn: Callable[..., AnthropicDynamicConfig | Awaitable[AnthropicDynamicConfig]],
    fn_args: dict[str, any],
    dynamic_config: AnthropicDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: AnthropicCallParams,
    extract: bool,
) -> tuple[
    Callable[..., Message],
    str,
    list[dict[str, MessageParam]],
    list[type[AnthropicTool]],
    dict[str, Any],
]:
    prompt_template, messages, tool_types, call_kwargs = _utils.setup_call(
        fn, fn_args, dynamic_config, tools, AnthropicTool, call_params
    )
    if messages[0]["role"] == "system":
        call_kwargs["system"] = messages.pop(0)["content"]
    if client is None:
        client = AsyncAnthropic() if inspect.iscoroutinefunction(fn) else Anthropic()
    create = client.messages.create
    call_kwargs |= {"model": model, "messages": messages}

    if json_mode:
        messages[-1]["content"] += _utils.json_mode_content(
            tool_types[0] if tools else None
        )
        call_kwargs.pop("tools", None)
    elif extract:
        call_kwargs["tool_choice"] = {"type": "tool", "name": tool_types[0]._name()}

    return create, prompt_template, messages, tool_types, call_kwargs


def get_json_output(response: Message, json_mode: bool):
    """Get the JSON output from a completion response."""
    block = response.content[0]
    if json_mode and block.type == "text":
        json_start = block.text.index("{")
        json_end = block.text.rfind("}")
        return block.text[json_start : json_end + 1]
    elif block.type == "tool_use":
        return block.input
    else:
        raise ValueError("No tool call or JSON object found in response.")


def _handle_chunk(
    buffer: str,
    chunk: MessageStreamEvent,
    current_tool_call: ToolUseBlock,
    current_tool_type: type[AnthropicTool] | None,
    tool_types: list[type[AnthropicTool]] | None,
) -> tuple[
    str,
    AnthropicTool | None,
    ToolUseBlock,
    type[AnthropicTool] | None,
]:
    """Handles a chunk of the stream."""
    if not tool_types:
        return buffer, None, current_tool_call, current_tool_type

    if chunk.type == "content_block_stop" and current_tool_type and buffer:
        current_tool_call.input = json.loads(buffer)
        return (
            "",
            current_tool_type.from_tool_call(current_tool_call),
            ToolUseBlock(id="", input={}, name="", type="tool_use"),
            None,
        )

    if chunk.type == "content_block_start" and isinstance(
        chunk.content_block, ToolUseBlock
    ):
        content_block = chunk.content_block
        current_tool_type = None
        for tool_type in tool_types:
            if tool_type._name() == content_block.name:
                current_tool_type = tool_type
                break
        if current_tool_type is None:
            raise RuntimeError(f"Unknown tool type in stream: {content_block.name}.")
        return (
            "",
            None,
            ToolUseBlock(
                id=content_block.id, input={}, name=content_block.name, type="tool_use"
            ),
            current_tool_type,
        )

    if chunk.type == "content_block_delta" and chunk.delta.type == "input_json_delta":
        buffer += chunk.delta.partial_json

    return buffer, None, current_tool_call, current_tool_type


def handle_stream(
    stream: Generator[MessageStreamEvent, None, None],
    tool_types: list[type[AnthropicTool] | Callable],
):
    """Iterator over the stream and constructs tools as they are streamed."""
    current_text_block = TextBlock(id="", text="", type="text")
    current_tool_call = ToolUseBlock(id="", input={}, name="", type="tool_use")
    current_tool_type, content, buffer = None, [], ""
    for chunk in stream:
        buffer, tool, current_tool_call, current_tool_type = _handle_chunk(
            buffer, chunk, current_tool_call, current_tool_type, tool_types
        )
        if tool is not None:
            yield chunk, tool
            if current_text_block.text:
                content.append(current_text_block)
                current_text_block = TextBlock(id="", text="", type="text")
            content.append(tool.tool_call)
        else:
            yield chunk, None
            if hasattr(chunk, "content_block") and hasattr(chunk.content_block, "text"):
                current_text_block.text += chunk.content_block.text


async def handle_stream_async(
    stream: AsyncGenerator[MessageStreamEvent, None],
    tool_types: list[type[AnthropicTool] | Callable],
):
    """Async iterator over the stream and constructs tools as they are streamed."""
    current_text_block = TextBlock(id="", text="", type="text")
    current_tool_call = ToolUseBlock(id="", input={}, name="", type="tool_use")
    current_tool_type, content, buffer = None, [], ""
    async for chunk in stream:
        buffer, tool, current_tool_call, current_tool_type = _handle_chunk(
            buffer, chunk, current_tool_call, current_tool_type, tool_types
        )
        if tool is not None:
            yield chunk, tool
            if current_text_block.text:
                content.append(current_text_block)
                current_text_block = TextBlock(id="", text="", type="text")
            content.append(tool.tool_call)
        else:
            yield chunk, None
            if hasattr(chunk, "content_block") and hasattr(chunk.content_block, "text"):
                current_text_block.text += chunk.content_block.text


def anthropic_api_calculate_cost(
    response: Message, model="claude-3-haiku-20240229"
) -> float | None:
    """Calculate the cost of a completion using the Anthropic API.

    https://www.anthropic.com/api

    claude-instant-1.2        $0.80 / 1M tokens   $2.40 / 1M tokens
    claude-2.0                $8.00 / 1M tokens   $24.00 / 1M tokens
    claude-2.1                $8.00 / 1M tokens   $24.00 / 1M tokens
    claude-3-haiku            $0.25 / 1M tokens   $1.25 / 1M tokens
    claude-3-sonnet           $3.00 / 1M tokens   $15.00 / 1M tokens
    claude-3-opus             $15.00 / 1M tokens   $75.00 / 1M tokens
    """
    pricing = {
        "claude-instant-1.2": {
            "prompt": 0.000_000_8,
            "completion": 0.000_002_4,
        },
        "claude-2.0": {
            "prompt": 0.000_008,
            "completion": 0.000_024,
        },
        "claude-2.1": {
            "prompt": 0.000_008,
            "completion": 0.000_024,
        },
        "claude-3-haiku-20240307": {
            "prompt": 0.000_002_5,
            "completion": 0.000_012_5,
        },
        "claude-3-sonnet-20240229": {
            "prompt": 0.000_003,
            "completion": 0.000_015,
        },
        "claude-3-opus-20240229": {
            "prompt": 0.000_015,
            "completion": 0.000_075,
        },
    }

    try:
        model = response.model if response.model else model
        model_pricing = pricing[model]
    except KeyError:
        return None

    prompt_cost = response.usage.input_tokens * model_pricing["prompt"]
    completion_cost = response.usage.output_tokens * model_pricing["completion"]
    total_cost = prompt_cost + completion_cost

    return total_cost



"""The Mirascope Gemini Module."""

from .call import gemini_call
from .call import gemini_call as call
from .call_async import gemini_call_async
from .call_async import gemini_call_async as call_async
from .call_params import GeminiCallParams
from .call_response import GeminiCallResponse
from .call_response_chunk import GeminiCallResponseChunk
from .dynamic_config import GeminiDynamicConfig
from .tool import GeminiTool

__all__ = [
    "call",
    "call_async",
    "GeminiDynamicConfig",
    "GeminiCallParams",
    "GeminiCallResponse",
    "GeminiCallResponseChunk",
    "GeminiTool",
    "gemini_call",
    "gemini_call_async",
]



"""Mirascope x OpenTelemetry Integration."""

import json
from contextlib import contextmanager
from typing import (
    Any,
    Awaitable,
    Callable,
    Generator,
    ParamSpec,
    Sequence,
    TypeVar,
    overload,
)

from opentelemetry.sdk.trace import SpanProcessor, TracerProvider
from opentelemetry.sdk.trace.export import (
    ConsoleSpanExporter,
    SimpleSpanProcessor,
)
from opentelemetry.trace import (
    Tracer,
    get_tracer,
    get_tracer_provider,
    set_tracer_provider,
)
from opentelemetry.trace.span import Span
from opentelemetry.util.types import Attributes, AttributeValue
from pydantic import BaseModel

from ..core.base import BaseAsyncStream, BaseCallResponse, BaseStream
from .utils import middleware

_P = ParamSpec("_P")
_R = TypeVar("_R", bound=BaseCallResponse | BaseStream | BaseModel | BaseAsyncStream)


def configure(
    processors: Sequence[SpanProcessor] | None = None,
) -> Tracer:
    """Configures the OpenTelemetry tracer, this function should only be called once.

    Args:
        processors: Optional[Sequence[SpanProcessor]]
            The span processors to use, if None, a console exporter will be used.

    Returns:
        The configured tracer.
    """
    provider = TracerProvider()
    if processors is None:
        provider.add_span_processor(SimpleSpanProcessor(ConsoleSpanExporter()))
    else:
        for processor in processors:
            provider.add_span_processor(processor)
    # NOTE: Sets the global trace provider, should only be called once
    set_tracer_provider(provider)
    return get_tracer("otel")


@overload
def with_otel(fn: Callable[_P, _R]) -> Callable[_P, _R]: ...


@overload
def with_otel(fn: Callable[_P, Awaitable[_R]]) -> Callable[_P, Awaitable[_R]]: ...


def with_otel(
    fn: Callable[_P, _R] | Callable[_P, Awaitable[_R]],
) -> Callable[_P, _R] | Callable[_P, Awaitable[_R]]:
    """Wraps a Mirascope function with OpenTelemetry tracing."""

    @contextmanager
    def custom_context_manager() -> Generator[Span, Any, None]:
        tracer = get_tracer("otel")
        with tracer.start_as_current_span(f"{fn.__name__}") as span:
            yield span

    def handle_call_response(result: BaseCallResponse, span: Span | None):
        if span is None:
            return
        max_tokens = getattr(result.call_params, "max_tokens", 0)
        temperature = getattr(result.call_params, "temperature", 0)
        top_p = getattr(result.call_params, "top_p", 0)
        attributes: dict[str, AttributeValue] = {
            "async": False,
            "gen_ai.system": result.prompt_template,
            "gen_ai.request.model": result.model,
            "gen_ai.request.max_tokens": max_tokens,
            "gen_ai.request.temperature": temperature,
            "gen_ai.request.top_p": top_p,
            "gen_ai.response.model": result.model,
            "gen_ai.response.id": result.id,
            "gen_ai.response.finish_reasons": result.finish_reasons,
            "gen_ai.usage.completion_tokens": result.output_tokens,
            "gen_ai.usage.prompt_tokens": result.input_tokens,
        }

        prompt_event: Attributes = {
            "gen_ai.prompt": json.dumps(result.user_message_param),
        }
        completion_event: Attributes = {
            "gen_ai.completion": json.dumps(result.message_param),
        }
        span.set_attributes(attributes)
        span.add_event("gen_ai.content.prompt", attributes=prompt_event)
        span.add_event(
            "gen_ai.content.completion",
            attributes=completion_event,
        )

    def handle_stream(stream: BaseStream, span: Span | None):
        if span is None:
            return
        max_tokens = getattr(stream.call_params, "max_tokens", 0)
        temperature = getattr(stream.call_params, "temperature", 0)
        top_p = getattr(stream.call_params, "top_p", 0)
        attributes: dict[str, AttributeValue] = {
            "async": False,
            "gen_ai.system": stream.prompt_template,
            "gen_ai.request.model": stream.model,
            "gen_ai.request.max_tokens": max_tokens,
            "gen_ai.request.temperature": temperature,
            "gen_ai.request.top_p": top_p,
            "gen_ai.response.model": stream.model,
            "gen_ai.usage.completion_tokens": stream.output_tokens,
            "gen_ai.usage.prompt_tokens": stream.input_tokens,
        }

        prompt_event: Attributes = {
            "gen_ai.prompt": json.dumps(stream.user_message_param),
        }
        completion_event: Attributes = {
            "gen_ai.completion": json.dumps(stream.message_param),
        }
        span.set_attributes(attributes)
        span.add_event("gen_ai.content.prompt", attributes=prompt_event)
        span.add_event(
            "gen_ai.content.completion",
            attributes=completion_event,
        )

    def handle_base_model(result: BaseModel, span: Span | None): ...

    async def handle_call_response_async(
        result: BaseCallResponse, span: Span | None
    ): ...

    async def handle_stream_async(stream: BaseStream, span: Span | None): ...

    provider = get_tracer_provider()
    if not isinstance(provider, TracerProvider):
        configure()
    return middleware(
        fn,
        custom_context_manager=custom_context_manager,
        handle_call_response=handle_call_response,
        handle_call_response_async=handle_call_response_async,
        handle_stream=handle_stream,
        handle_stream_async=handle_stream_async,
        handle_base_model=handle_base_model,
        # handle_base_model_async=handle_base_model_async,
    )



"""This module contains the Anthropic `structured_stream_async_decorator` function."""

from collections.abc import AsyncGenerator
from functools import wraps
from typing import (
    AsyncIterable,
    Awaitable,
    Callable,
    Generic,
    ParamSpec,
    TypeVar,
)

from anthropic import AsyncAnthropic
from anthropic.lib.streaming import MessageStreamEvent
from pydantic import BaseModel

from ..base import BaseAsyncStructuredStream, _utils

# from ._utils import setup_extract
from .call_params import AnthropicCallParams
from .dynamic_config import AnthropicDynamicConfig
from .tool import AnthropicTool

_P = ParamSpec("_P")
_ResponseModelT = TypeVar("_ResponseModelT", bound=BaseModel | _utils.BaseType)


class AnthropicAsyncStructuredStream(
    Generic[_ResponseModelT],
    BaseAsyncStructuredStream[MessageStreamEvent, _ResponseModelT],
):
    """A class for async streaming structured outputs from Anthropic's API."""

    def __aiter__(self) -> AsyncGenerator[_ResponseModelT, None]:
        """Iterates over the stream and extracts structured outputs."""

        async def generator():
            nonlocal self
            json_output = ""
            async for chunk in self.stream:
                if chunk.type == "content_block_delta":
                    if chunk.delta.type in ["text", "text_delta"]:
                        json_output += chunk.delta.text
                    elif chunk.delta.type == "input_json_delta":
                        json_output += chunk.delta.partial_json
                if json_output:
                    yield _utils.extract_tool_return(
                        self.response_model, json_output, True
                    )
            yield _utils.extract_tool_return(self.response_model, json_output, False)

        return generator()


def structured_stream_async_decorator(
    fn: Callable[_P, Awaitable[AnthropicDynamicConfig]],
    model: str,
    response_model: type[_ResponseModelT],
    call_params: AnthropicCallParams,
) -> Callable[_P, Awaitable[AsyncIterable[_ResponseModelT]]]:
    assert response_model is not None
    tool = _utils.setup_extract_tool(response_model, AnthropicTool)

    @wraps(fn)
    async def inner(
        *args: _P.args, **kwargs: _P.kwargs
    ) -> AsyncIterable[_ResponseModelT]:
        # assert response_model is not None
        # fn_args = _utils.get_fn_args(fn, args, kwargs)
        # fn_return = await fn(*args, **kwargs)
        # json_mode, messages, call_kwargs = setup_extract(
        #     fn, fn_args, fn_return, tool, call_params
        # )
        # client = AsyncAnthropic()
        # return AnthropicAsyncStructuredStream(
        #     stream=(
        #         chunk
        #         async for chunk in await client.messages.create(
        #             model=model, stream=True, messages=messages, **call_kwargs
        #         )
        #     ),
        #     response_model=response_model,
        #     json_mode=json_mode,
        # )
        ...

    return inner



"""The Mirascope Core Functionality."""

from contextlib import suppress

from . import base
from .base import BasePrompt, prompt_template, tags

with suppress(ImportError):
    from . import openai as openai

__all__ = ["base", "BasePrompt", "openai", "prompt_template", "tags"]



"""Mirascope Base Classes."""

from . import _partial, _utils
from .call_async_factory import call_async_factory
from .call_factory import call_factory
from .call_params import BaseCallParams
from .call_response import BaseCallResponse
from .call_response_chunk import BaseCallResponseChunk
from .dynamic_config import BaseDynamicConfig
from .message_param import BaseMessageParam
from .prompt import BasePrompt, prompt_template, tags
from .structured_stream import BaseStructuredStream
from .structured_stream_async import BaseAsyncStructuredStream
from .tool import BaseTool
from .toolkit import BaseToolKit, toolkit_tool

__all__ = [
    "BaseAsyncStructuredStream",
    "BaseCallParams",
    "BaseCallResponse",
    "BaseCallResponseChunk",
    "BaseDynamicConfig",
    "BaseMessageParam",
    "BasePrompt",
    "BaseStructuredStream",
    "BaseTool",
    "BaseToolKit",
    "call_async_factory",
    "call_factory",
    "prompt_template",
    "tags",
    "toolkit_tool",
    "_partial",
    "_utils",
]



"""The `openai_call_async` decorator for easy OpenAI API typed functions."""

from openai.types.chat import (
    ChatCompletionAssistantMessageParam,
    ChatCompletionMessageParam,
    ChatCompletionUserMessageParam,
)

from ..base import call_async_factory
from ..base._stream import BaseStream
from ._structured_stream_async import structured_stream_async_decorator
from ._utils import (
    get_json_output,
    handle_stream,
    handle_stream_async,
    openai_api_calculate_cost,
    setup_call,
)
from .call_params import OpenAICallParams
from .call_response import OpenAICallResponse
from .call_response_chunk import OpenAICallResponseChunk
from .dyanmic_config import OpenAIDynamicConfig
from .tool import OpenAITool

OpenAIAsyncStream = BaseStream[
    OpenAICallResponseChunk,
    ChatCompletionUserMessageParam,
    ChatCompletionAssistantMessageParam,
    ChatCompletionMessageParam,
    OpenAITool,
    OpenAIDynamicConfig,
]


openai_call_async = call_async_factory(
    TCallResponse=OpenAICallResponse,
    TCallResponseChunk=OpenAICallResponseChunk,
    TCallParams=OpenAICallParams,
    TDynamicConfig=OpenAIDynamicConfig,
    TStream=OpenAIAsyncStream,
    TMessageParamType=ChatCompletionAssistantMessageParam,
    TToolType=OpenAITool,
    setup_call=setup_call,
    get_json_output=get_json_output,
    handle_stream=handle_stream,
    handle_stream_async=handle_stream_async,
    calculate_cost=openai_api_calculate_cost,
    structured_stream_async_decorator=structured_stream_async_decorator,
)
'''A decorator for calling the AsyncOpenAI API with a typed function.

This decorator is used to wrap a typed function that calls the OpenAI API. It parses
the docstring of the wrapped function as the messages array and templates the input
arguments for the function into each message's template.

Example:

```python
@openai_call_async(model="gpt-4o")
async def recommend_book(genre: str):
    """Recommend a {genre} book."""

async def run():
    response = await recommend_book("fantasy")

asyncio.run(run())
```

Args:
    model: The OpenAI model to use in the API call.
    stream: Whether to stream the response from the API call.
    tools: The tools to use in the OpenAI API call.
    **call_params: The `OpenAICallParams` call parameters to use in the API call.

Returns:
    The decorator for turning a typed function into an AsyncOpenAI API call.
'''



"""This module contains the base class for message parameters."""

from typing import Literal

from typing_extensions import TypedDict


class BaseMessageParam(TypedDict):
    """A base class for message parameters.

    Available roles: `system`, `user`, `assistant`.
    """

    role: Literal["system", "user", "assistant"]
    content: str



"""This module defines the base class for tools used in LLM calls."""

import inspect
from abc import abstractmethod
from typing import Any

from pydantic import BaseModel, ConfigDict

from . import _utils


class BaseTool(BaseModel):
    """A class for defining tools for LLM calls."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @classmethod
    def _name(cls) -> str:
        """Returns the name of the tool."""
        return cls.__name__

    @classmethod
    def _description(cls) -> str:
        """Returns the description of the tool."""
        return (
            inspect.cleandoc(cls.__doc__)
            if cls.__doc__
            else _utils.DEFAULT_TOOL_DOCSTRING
        )

    @abstractmethod
    def call(self) -> Any:
        """The method to call the tool."""
        ...  # pragma: no cover



"""This module contains the base classes for structured streams from LLMs."""

from abc import ABC, abstractmethod
from collections.abc import Generator
from typing import Any, Generic, TypeVar

from pydantic import BaseModel

from ._utils import BaseType

_ChunkT = TypeVar("_ChunkT", bound=Any)
_ResponseModelT = TypeVar("_ResponseModelT", bound=BaseModel | BaseType)


class BaseStructuredStream(Generic[_ChunkT, _ResponseModelT], ABC):
    """A base class for streaming structured outputs from LLMs."""

    stream: Generator[_ChunkT, None, None]
    response_model: type[_ResponseModelT]
    json_mode: bool

    def __init__(
        self,
        stream: Generator[_ChunkT, None, None],
        response_model: type[_ResponseModelT],
        json_mode: bool = False,
    ):
        """Initializes an instance of `BaseStructuredStream`."""
        self.stream = stream
        self.response_model = response_model
        self.json_mode = json_mode

    @abstractmethod
    def __iter__(self) -> Generator[_ResponseModelT, None, None]:
        """Iterates over the stream and extracts structured outputs."""



"""This module contains the Gemini `structured_stream_async_decorator` function."""

import json
from collections.abc import AsyncGenerator
from functools import wraps
from typing import AsyncIterable, Awaitable, Callable, Generic, ParamSpec, TypeVar

from google.generativeai import GenerativeModel  # type: ignore
from google.generativeai.types import GenerateContentResponse
from pydantic import BaseModel

from ..base import BaseAsyncStructuredStream, _utils

# from ._utils import setup_extract
from .call_params import GeminiCallParams
from .dynamic_config import GeminiDynamicConfig
from .tool import GeminiTool

_P = ParamSpec("_P")
_ResponseModelT = TypeVar("_ResponseModelT", bound=BaseModel | _utils.BaseType)


class GeminiAsyncStructuredStream(
    Generic[_ResponseModelT],
    BaseAsyncStructuredStream[GenerateContentResponse, _ResponseModelT],
):
    """A class for streaming structured outputs from Gemini's API."""

    def __aiter__(self) -> AsyncGenerator[_ResponseModelT, None]:
        """Iterates over the stream and extracts structured outputs."""

        async def generator():
            nonlocal self
            json_output = ""
            async for chunk in self.stream:
                if (
                    self.json_mode
                    and (content := chunk.candidates[0].content.parts[0].text)
                    is not None
                ):
                    print(content)
                    json_output += content
                elif (
                    tool_calls := chunk.candidates[0].content.parts[0].function_call
                ) and (arguments := tool_calls.args):
                    json_output += json.dumps(dict(arguments))
                else:
                    ValueError("No tool call or JSON object found in response.")
                if json_output:
                    yield _utils.extract_tool_return(
                        self.response_model, json_output, True
                    )
            yield _utils.extract_tool_return(self.response_model, json_output, False)

        return generator()


def structured_stream_async_decorator(
    fn: Callable[_P, Awaitable[GeminiDynamicConfig]],
    model: str,
    response_model: type[_ResponseModelT],
    call_params: GeminiCallParams,
) -> Callable[_P, Awaitable[AsyncIterable[_ResponseModelT]]]:
    assert response_model is not None
    tool = _utils.setup_extract_tool(response_model, GeminiTool)

    @wraps(fn)
    async def inner_async(
        *args: _P.args, **kwargs: _P.kwargs
    ) -> AsyncIterable[_ResponseModelT]:
        # assert response_model is not None
        # fn_args = _utils.get_fn_args(fn, args, kwargs)
        # fn_return = await fn(*args, **kwargs)
        # json_mode, messages, call_kwargs = setup_extract(
        #     fn, fn_args, fn_return, tool, call_params
        # )
        # client = GenerativeModel(model_name=model)
        # stream = client.generate_content_async(
        #     messages,
        #     stream=True,
        #     **call_kwargs,
        # )
        # return GeminiAsyncStructuredStream(
        #     stream=stream,
        #     response_model=response_model,
        #     json_mode=json_mode,
        # )
        ...

    return inner_async



"""This module contains the `get_template_values` function."""

from typing import Any


def get_template_values(
    template_variables: list[str], attrs: dict[str, Any]
) -> dict[str, Any]:
    """Returns the values of the given `template_variables` from the provided `attrs`.

    Args:
        template_variables: The variables to extract from the `attrs`.
        attrs: The attributes to extract the variables from.

    Returns:
        The values of the template variables.
    """
    values = {}
    if "self" in attrs:
        values["self"] = attrs.get("self")
    for var in template_variables:
        if var.startswith("self"):
            continue
        if "." in var:
            var = var.split(".")[0]
            values[var] = attrs.get(var)
            print(values[var])
            continue
        attr = attrs[var]
        if isinstance(attr, list):
            if len(attr) == 0:
                values[var] = ""
            elif isinstance(attr[0], list):
                values[var] = "\n\n".join(
                    ["\n".join([str(subitem) for subitem in item]) for item in attr]
                )
            else:
                values[var] = "\n".join([str(item) for item in attr])
        else:
            values[var] = str(attr)
    return values



"""The Mirascope Anthropic Module."""

from .call import anthropic_call
from .call import anthropic_call as call
from .call_async import anthropic_call_async
from .call_async import anthropic_call_async as call_async
from .call_params import AnthropicCallParams
from .call_response import AnthropicCallResponse
from .call_response_chunk import AnthropicCallResponseChunk
from .dynamic_config import AnthropicDynamicConfig
from .tool import AnthropicTool

__all__ = [
    "call",
    "call_async",
    "AnthropicDynamicConfig",
    "AnthropicCallParams",
    "AnthropicCallResponse",
    "AnthropicCallResponseChunk",
    "AnthropicTool",
    "anthropic_call",
    "anthropic_call_async",
]



"""This module contains the base classes for streaming responses from LLMs."""

import inspect
from abc import ABC
from collections.abc import AsyncGenerator, Generator
from functools import wraps
from typing import Any, Awaitable, Callable, Generic, ParamSpec, TypeVar, overload

from ._utils import get_fn_args
from .call_params import BaseCallParams
from .call_response_chunk import BaseCallResponseChunk
from .dynamic_config import BaseDynamicConfig
from .tool import BaseTool

_BaseCallResponseChunkT = TypeVar(
    "_BaseCallResponseChunkT", bound=BaseCallResponseChunk
)
_UserMessageParamT = TypeVar("_UserMessageParamT")
_AssistantMessageParamT = TypeVar("_AssistantMessageParamT")
_MessageParamT = TypeVar("_MessageParamT")
_BaseToolT = TypeVar("_BaseToolT", bound=BaseTool)
_CallParamsT = TypeVar("_CallParamsT", bound=BaseCallParams)
_BaseDynamicConfigT = TypeVar("_BaseDynamicConfigT", bound=BaseDynamicConfig)


class BaseStream(
    Generic[
        _BaseCallResponseChunkT,
        _UserMessageParamT,
        _AssistantMessageParamT,
        _MessageParamT,
        _BaseToolT,
        _BaseDynamicConfigT,
    ],
    ABC,
):
    """A base class for streaming responses from LLMs."""

    stream: (
        Generator[tuple[_BaseCallResponseChunkT, _BaseToolT | None], None, None]
        | AsyncGenerator[
            tuple[_BaseCallResponseChunkT, _BaseToolT | None],
            None,
        ]
    )
    tags: list[str]
    tools_types: list[type[_BaseToolT]] | None
    message_param_type: type[_AssistantMessageParamT]
    model: str | None = None
    cost: float | None = None
    prompt_template: str | None = None
    fn_args: dict[str, Any] | None = None
    dynamic_config: _BaseDynamicConfigT
    messages: list[_MessageParamT]
    call_params: _CallParamsT
    input_tokens: int | float | None = None
    output_tokens: int | float | None = None
    user_message_param: _UserMessageParamT | None = None
    message_param: _AssistantMessageParamT
    provider: str

    def __init__(
        self,
        stream: Generator[tuple[_BaseCallResponseChunkT, _BaseToolT | None], None, None]
        | AsyncGenerator[
            tuple[_BaseCallResponseChunkT, _BaseToolT | None],
            None,
        ],
        *,
        tags: list[str],
        tool_types: list[type[_BaseToolT]] | None,
        message_param_type: type[_AssistantMessageParamT],
        model: str,
        cost: float | None,
        prompt_template: str,
        fn_args: dict[str, Any],
        dynamic_config: _BaseDynamicConfigT,
        messages: list[_MessageParamT],
        call_params: _CallParamsT,
        user_message_param: _UserMessageParamT | None,
    ):
        """Initializes an instance of `BaseStream`."""
        self.stream = stream
        self.tags = tags
        self.tool_types = tool_types
        self.message_param_type = message_param_type
        self.model = model
        self.cost = cost
        self.prompt_template = prompt_template
        self.fn_args = fn_args
        self.dynamic_config = dynamic_config
        self.messages = messages
        self.call_params = call_params
        self.user_message_param = user_message_param

    def __iter__(
        self,
    ) -> Generator[tuple[_BaseCallResponseChunkT, _BaseToolT | None], None, None]:
        """Iterator over the stream and stores useful information."""
        assert isinstance(
            self.stream, Generator
        ), "Stream must be a generator for __iter__"
        content, tool_calls = "", []
        for chunk, tool in self.stream:
            content += chunk.content
            if chunk.input_tokens is not None:
                self.input_tokens = chunk.input_tokens
            if chunk.output_tokens is not None:
                self.output_tokens = chunk.output_tokens
            if chunk.model is not None:
                self.model = chunk.model
            if tool:
                tool_calls.append(tool.tool_call)  # type: ignore
            yield chunk, tool
        kwargs = {"role": "assistant"}
        if "message" in self.message_param_type.__annotations__:
            kwargs["message"] = content
        else:
            kwargs["content"] = content
        if tool_calls:
            kwargs["tool_calls"] = tool_calls
        self.message_param = self.message_param_type(**kwargs)

    def __aiter__(
        self,
    ) -> AsyncGenerator[tuple[_BaseCallResponseChunkT, _BaseToolT | None], None]:
        """Iterates over the stream and stores useful information."""

        async def generator():
            assert isinstance(
                self.stream, AsyncGenerator
            ), "Stream must be an async generator for __aiter__"
            content, tool_calls = "", []
            async for chunk, tool in self.stream:
                content += chunk.content
                if chunk.input_tokens is not None:
                    self.input_tokens = chunk.input_tokens
                if chunk.output_tokens is not None:
                    self.output_tokens = chunk.output_tokens
                if chunk.model is not None:
                    self.model = chunk.model
                if tool:
                    tool_calls.append(tool.tool_call)  # type: ignore
                yield chunk, None
            kwargs = {"role": "assistant"}
            if "message" in self.message_param_type.__annotations__:
                kwargs["message"] = content
                if tool_calls:
                    kwargs["tool_calls"] = tool_calls
            else:
                kwargs["content"] = content
                if tool_calls:
                    kwargs["content"] = [kwargs["content"]] + tool_calls
            self.message_param = self.message_param_type(**kwargs)

        return generator()


_BaseCallResponseChunkT = TypeVar(
    "_BaseCallResponseChunkT", bound=BaseCallResponseChunk
)
_BaseDynamicConfigT = TypeVar("_BaseDynamicConfigT", bound=BaseDynamicConfig)
_BaseClientT = TypeVar("_BaseClientT", bound=object)
_BaseCallParamsT = TypeVar("_BaseCallParamsT", bound=BaseCallParams)
_ResponseT = TypeVar("_ResponseT")
_P = ParamSpec("_P")


def stream_factory(
    TCallResponseChunk: type[_BaseCallResponseChunkT],
    TStream: type[BaseStream],
    TMessageParamType: type[_AssistantMessageParamT],
    setup_call: Callable[
        [
            str,
            _BaseClientT,
            Callable[_P, _BaseDynamicConfigT | Awaitable[_BaseDynamicConfigT]],
            dict[str, Any],
            _BaseDynamicConfigT,
            list[type[BaseTool] | Callable] | None,
            _BaseCallParamsT,
            bool,
        ],
        tuple[
            Callable[..., _ResponseT],
            str,
            list[dict[str, Any]],
            list[type[BaseTool]],
            dict[str, Any],
        ],
    ],
    handle_stream: Callable[
        [Generator[_BaseCallResponseChunkT, None, None], list[type[_BaseToolT]]],
        Generator[tuple[_BaseCallResponseChunkT, _BaseToolT], None, None],
    ],
    handle_stream_async: Callable[
        [AsyncGenerator[_BaseCallResponseChunkT, None], list[type[_BaseToolT]]],
        Awaitable[AsyncGenerator[tuple[_BaseCallResponseChunkT, _BaseToolT], None]],
    ],
):
    @overload
    def decorator(
        fn: Callable[_P, _BaseDynamicConfigT],
        model: str,
        tools: list[type[BaseTool] | Callable] | None,
        json_mode: bool,
        client: _BaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[_P, TStream]: ...  # pragma: no cover

    @overload
    def decorator(
        fn: Callable[_P, Awaitable[_BaseDynamicConfigT]],
        model: str,
        tools: list[type[BaseTool] | Callable] | None,
        json_mode: bool,
        client: _BaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[_P, Awaitable[TStream]]: ...  # pragma: no cover

    def decorator(
        fn: Callable[_P, _BaseDynamicConfigT],
        model: str,
        tools: list[type[BaseTool] | Callable] | None,
        json_mode: bool,
        client: _BaseClientT | None,
        call_params: _BaseCallParamsT,
    ) -> Callable[_P, TStream | Awaitable[TStream]]:
        is_async = inspect.iscoroutinefunction(fn)

        @wraps(fn)
        def inner(*args: _P.args, **kwargs: _P.kwargs) -> TStream:
            fn_args = get_fn_args(fn, args, kwargs)
            dynamic_config = fn(*args, **kwargs)
            create, prompt_template, messages, tool_types, call_kwargs = setup_call(
                model=model,
                client=client,
                fn=fn,
                fn_args=fn_args,
                dynamic_config=dynamic_config,
                tools=tools,
                json_mode=json_mode,
                call_params=call_params,
                extract=False,
            )

            def generator():
                for chunk, tool in handle_stream(
                    create(stream=True, **call_kwargs), tool_types
                ):
                    yield TCallResponseChunk(chunk=chunk), tool

            return TStream(
                generator(),
                tags=fn.__annotations__.get("tags", []),
                tool_types=tool_types,
                message_param_type=TMessageParamType,
                model=model,
                cost=None,
                prompt_template=prompt_template,
                fn_args=fn_args,
                dynamic_config=dynamic_config,
                messages=messages,
                call_params=call_params,
                user_message_param=messages[-1]
                if messages[-1]["role"] == "user"
                else None,
            )

        @wraps(fn)
        async def inner_async(*args: _P.args, **kwargs: _P.kwargs) -> TStream:
            fn_args = get_fn_args(fn, args, kwargs)
            dynamic_config = await fn(*args, **kwargs)
            create, prompt_template, messages, tool_types, call_kwargs = setup_call(
                model=model,
                client=client,
                fn=fn,
                fn_args=fn_args,
                dynamic_config=dynamic_config,
                tools=tools,
                json_mode=json_mode,
                call_params=call_params,
                extract=False,
            )

            async def generator():
                async for chunk, tool in handle_stream_async(
                    await create(stream=True, **call_kwargs), tool_types
                ):
                    yield TCallResponseChunk(chunk=chunk), tool

            return TStream(
                generator(),
                tags=fn.__annotations__.get("tags", []),
                tool_types=tool_types,
                message_param_type=TMessageParamType,
                model=model,
                cost=None,
                prompt_template=prompt_template,
                fn_args=fn_args,
                dynamic_config=dynamic_config,
                messages=messages,
                call_params=call_params,
                user_message_param=messages[-1]
                if messages[-1]["role"] == "user"
                else None,
            )

        return inner_async if is_async else inner

    return decorator
